@@ -1,6 +1,6 @@
 # ============================================================
 #
-# Copyright (C) 2011, 2012 Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
+# Copyright (C) 2011, 2012, 2014 Jan Moringen <jmoringe@techfak.uni-bielefeld.de>
 #
 # This file may be licensed under the terms of the
 # GNU Lesser General Public License Version 3 (the ``LGPL''),
@@ -34,6 +34,8 @@ import uuid
 import threading
 
 import rsb
+import rsb.filter
+
 from future import Future, DataFuture
 from rsb.eventprocessing import FullyParallelEventReceivingStrategy
 
@@ -74,7 +76,7 @@ class RemoteCallError (RuntimeError):
 #
 ######################################################################
 
-class Method (object):
+class Method (rsb.Participant):
     """
     Objects of this class are methods which are associated to a local
     or remote server. Within a server, each method has a unique name.
@@ -84,14 +86,16 @@ class Method (object):
 
     @author: jmoringe
     """
-    def __init__(self, server, name, requestType, replyType):
+    def __init__(self, scope, config,
+                 server, name, requestType, replyType): # TODO scope and name are redundant
         """
-        Create a new Method object for the method named B{name}
+        Create a new L{Method} object for the method named B{name}
         provided by B{server}.
 
         @param server: The remote or local server to which the method
                        is associated.
         @param name: The name of the method. Unique within a server.
+        @type  name: str
         @param requestType: The type of the request argument accepted
                             by the method.
         @type  requestType: type
@@ -99,6 +103,8 @@ class Method (object):
                           method.
         @type  replyType: type
         """
+        super(Method, self).__init__(scope, config)
+
         self._server      = server
         self._name        = name
         self._listener    = None
@@ -148,6 +154,8 @@ class Method (object):
             self._listener.deactivate()
             self._listener = None
 
+        super(Method, self).deactivate()
+
     def __str__(self):
         return '<%s "%s" at 0x%x>' % (type(self).__name__, self.name, id(self))
 
@@ -166,24 +174,21 @@ class Server (rsb.Participant):
     @author: jmoringe
     """
 
-    def __init__(self, scope, config = None):
+    def __init__(self, scope, config):
         """
-        Create a new Server object that provides its methods under the
-        scope B{scope}.
+        Create a new L{Server} object that provides its methods under the
+        L{rsb.Scope} B{scope}.
 
         @param scope: The under which methods of the server are
                       provided.
+        @type  scope: rsb.Scope
         @param config: The transport configuration that should be used
                        for communication performed by this server.
-        @type config: ParticipantConfig
+        @type  config: rsb.ParticipantConfig
         """
-        super(Server, self).__init__(scope)
+        super(Server, self).__init__(scope, config)
 
         self.__active = False
-        if config is None:
-            self.__config = rsb.getDefaultParticipantConfig()
-        else:
-            self.__config = config
         self._methods = {}
 
         self.activate()
@@ -192,11 +197,6 @@ class Server (rsb.Participant):
         if self.__active:
             self.deactivate()
 
-    def getConfig(self):
-        return self.__config
-
-    config = property(getConfig)
-
     def getMethods(self):
         return self._methods.values()
 
@@ -217,6 +217,8 @@ class Server (rsb.Participant):
     def activate(self):
         self.__active = True
 
+        super(Server, self).activate()
+
     def deactivate(self):
         if not self.__active:
             raise RuntimeError, 'Trying to deactivate inactive server'
@@ -226,6 +228,8 @@ class Server (rsb.Participant):
         for m in self._methods.values():
             m.deactivate()
 
+        super(Server, self).deactivate()
+
     # Printing
 
     def __str__(self):
@@ -250,8 +254,10 @@ class LocalMethod (Method):
 
     @author: jmoringe
     """
-    def __init__(self, server, name, func, requestType, replyType, allowParallelExecution):
-        super(LocalMethod, self).__init__(server, name, requestType, replyType)
+    def __init__(self, scope, config,
+                 server, name, func, requestType, replyType, allowParallelExecution):
+        super(LocalMethod, self).__init__(scope, config, server, name, requestType, replyType)
+
         self._allowParallelExecution = allowParallelExecution
         self._func = func
         self.listener # force listener creation
@@ -260,26 +266,19 @@ class LocalMethod (Method):
         receivingStrategy = None
         if self._allowParallelExecution:
             receivingStrategy = FullyParallelEventReceivingStrategy()
-        listener = rsb.Listener(self.server.scope
-                                .concat(rsb.Scope("/request"))
-                                .concat(rsb.Scope('/' + self.name)),
-                                self.server.config,
-                                receivingStrategy = receivingStrategy)
+        listener = rsb.createListener(self.scope, self.config,
+                                      parent            = self,
+                                      receivingStrategy = receivingStrategy)
+        listener.addFilter(rsb.filter.MethodFilter(method =  'REQUEST'))
         listener.addHandler(self._handleRequest)
         return listener
 
     def makeInformer(self):
-        return rsb.Informer(self.server.scope
-                            .concat(rsb.Scope("/reply"))
-                            .concat(rsb.Scope('/' + self.name)),
-                            object,
-                            config = self.server.config)
+        return rsb.createInformer(self.scope, self.config,
+                                  parent   = self,
+                                  dataType = object)
 
     def _handleRequest(self, request):
-        # Only accept request, if its method is 'REQUEST'.
-        if request.method is None or request.method != 'REQUEST':
-            return
-
         # Call the callable implementing the behavior of this
         # method. If it does not take an argument
         # (i.e. self.requestType is type(None)), call it without
@@ -292,6 +291,8 @@ class LocalMethod (Method):
             if self.requestType is type(None):
                 assert(request.data is None)
                 result = self._func()
+            elif self.requestType is rsb.Event:
+                result = self._func(request)
             else:
                 result = self._func(request.data)
             resultType = type(result)
@@ -306,6 +307,7 @@ class LocalMethod (Method):
         # necessary meta-data.
         if isinstance(result, rsb.Event):
             reply = result
+            reply.method = 'REPLY'
             reply.causes += causes
         else:
             # This check is required because the reply informer is
@@ -331,19 +333,19 @@ class LocalServer (Server):
 
     @author: jmoringe
     """
-    def __init__(self, scope, config = None):
+    def __init__(self, scope, config):
         """
         Creates a new L{LocalServer} object that exposes methods under
-        a the scope B{scope}.
+        the L{rsb.Scope} B{scope}.
 
         @param scope: The scope under which the methods of the newly
                       created server should be provided.
-        @type scope: Scope
+        @type scope: rsb.Scope
         @param config: The transport configuration that should be used
-        for communication performed by this server.
-        @type config: ParticipantConfig
+                       for communication performed by this server.
+        @type config: rsb.ParticipantConfig
 
-        See also: L{createServer}
+        See also: L{rsb.createServer}
         """
         super(LocalServer, self).__init__(scope, config)
 
@@ -372,11 +374,18 @@ class LocalServer (Server):
         @return: The newly created method.
         @rtype: LocalMethod
         """
-        method = LocalMethod(self, name, func, requestType, replyType,
-                             allowParallelExecution)
+        scope = self.scope.concat(rsb.Scope('/' + name))
+        method = rsb.createParticipant(LocalMethod, scope, self.config,
+                                       parent                 = self,
+                                       server                 = self,
+                                       name                   = name,
+                                       func                   = func,
+                                       requestType            = requestType,
+                                       replyType              = replyType,
+                                       allowParallelExecution = allowParallelExecution)
         super(LocalServer, self).addMethod(method)
         return method
-    
+
     def removeMethod(self, method):
         if isinstance(method, str):
             method = self.getMethod(method)
@@ -396,31 +405,29 @@ class RemoteMethod (Method):
 
     @author: jmoringe
     """
-    def __init__(self, server, name, requestType, replyType):
-        super(RemoteMethod, self).__init__(server, name, requestType, replyType)
+    def __init__(self, scope, config, server, name, requestType, replyType):
+        super(RemoteMethod, self).__init__(scope, config,
+                                           server, name, requestType, replyType)
+
         self._calls = {}
         self._lock  = threading.RLock()
 
     def makeListener(self):
-        listener = rsb.Listener(self.server.scope
-                                .concat(rsb.Scope("/reply"))
-                                .concat(rsb.Scope('/' + self.name)),
-                                self.server.config)
+        listener = rsb.createListener(self.scope, self.config,
+                                      parent = self)
+        listener.addFilter(rsb.filter.MethodFilter(method = 'REPLY'))
         listener.addHandler(self._handleReply)
         return listener
 
     def makeInformer(self):
-        return rsb.Informer(self.server.scope
-                            .concat(rsb.Scope("/request"))
-                            .concat(rsb.Scope('/' + self.name)),
-                            self.requestType,
-                            config = self.server.config)
+        return rsb.createInformer(self.scope, self.config,
+                                  parent   = self,
+                                  dataType = self.requestType)
 
     def _handleReply(self, event):
-        if event.method is None            \
-                or event.method != 'REPLY' \
-                or not event.causes:
+        if not event.causes:
             return
+
         key = event.causes[0]
         with self._lock:
             # We can receive reply events which aren't actually
@@ -541,28 +548,40 @@ class RemoteServer (Server):
 
     @author: jmoringe
     """
-    def __init__(self, scope, config = None):
+    def __init__(self, scope, config):
         """
         Create a new L{RemoteServer} object that provides its methods
         under the scope B{scope}.
 
         @param scope: The common super-scope under which the methods
-        of the remote created server are provided.
-        @type scope: Scope
-        @param config: The transport configuration that should be used
-        for communication performed by this server.
-        @type config: ParticipantConfig
-        @see: L{createRemoteServer}
+                      of the remote created server are provided.
+        @type scope: rsb.Scope
+        @param config: The configuration that should be used by this
+                       server.
+        @type config: rsb.ParticipantConfig
+        @see: L{rsb.createRemoteServer}
         """
         super(RemoteServer, self).__init__(scope, config)
 
+    def ensureMethod(self, name):
+        method = super(RemoteServer, self).getMethod(name)
+        if method is None:
+            scope = self.scope.concat(rsb.Scope('/' + name))
+            method = rsb.createParticipant(RemoteMethod, scope, self.config,
+                                           parent      = self,
+                                           server      = self,
+                                           name        = name,
+                                           requestType = object,
+                                           replyType   = object)
+            self.addMethod(method)
+        return method
+
+    def getMethod(self, name):
+        return self.ensureMethod(name)
+
     def __getattr__(self, name):
         # Treat missing attributes as methods.
         try:
-            super(RemoteServer, self).__getattr__(name)
+            return super(RemoteServer, self).__getattr__(name)
         except AttributeError:
-            method = self.getMethod(name)
-            if method is None:
-                method = RemoteMethod(self, name, object, object)
-                self.addMethod(method)
-            return method
+            return self.ensureMethod(name)
